'##################################################################################################
Type TWebServer  Extends TModule
	
	'################################################
	Field Socket:TSocket
	Field Func:Int( S:TSession )
	
	Field ThreadList:TList = CreateList()
	Field ThreadListMutex:TMutex = CreateMutex()
	Field ThreadCount:Int
	
	'################################################
	Function Create:TWebServer( Port:Int = 80, Func:Int( S:TSession ))
		Local A:TWebServer = New TWebServer
		A.LocalCreate( Port )
		A.Func = Func
		Return( A )
	End Function
	
	'################################################
	Method LocalCreate( Port:Int = 80 )
		Socket = CreateTCPSocket()
		If( Socket )
			Self.Log( "Socket.Create()", LOGLVL_INFO )
			If( BindSocket( Socket, Port ) )
				Self.Log( "Socket.Bind( "+Port+" )", LOGLVL_INFO )
				If( SocketListen( Socket, 5 ) )
					Self.Log( "Socket.Listen()", LOGLVL_INFO )
				Else
					Self.Log( "<<ERROR>> Socket.Listen() Failed!", LOGLVL_ERROR )
				EndIf
			Else
				Self.Log( "<<ERROR>> Socket.Bind( "+Port+" ) Failed!", LOGLVL_ERROR )
			EndIf
		Else
			Self.Log( "<<ERROR>> Socket.Create() Failed!", LOGLVL_ERROR )
		EndIf
	End Method
	
	'################################################
	Method Update()
		Local Accept:TSocket = SocketAccept( Socket )
		If( Accept )
			
			Local Thread:TThread = CreateThread( Self.NewConnection, TWebServerThreadEntry.Create( Accept, Self ) )
			ListAddLast( ThreadList, Thread )
			ThreadCount = ThreadCount + 1
			Self.Log( "Thread created! (Up to " + ThreadCount + " threads)", LOGLVL_DEBUG )
			
		EndIf
		
		For Local Thread:TThread = EachIn ThreadList
			If Not ThreadRunning( Thread )
				DetachThread( Thread )
				ListRemove( ThreadList, Thread )
				ThreadCount = ThreadCount - 1
				Self.Log( "Thread removed! (down to " + ThreadCount + " threads)", LOGLVL_DEBUG )
			EndIf
		Next
	End Method
	
	'################################################
	Method Kill()
		LockMutex( ThreadListMutex )
			For Local Thread:TThread = EachIn ThreadList
				WaitThread( Thread )
			Next
		UnlockMutex( ThreadListMutex )
	End Method
	
	'################################################
	Function NewConnection:Object( E:Object )
		Local Entry:TWebServerThreadEntry = TWebServerThreadEntry( E )
		
		Local A:TSession = New TSession
		A.Connection = TConnection.Create( Entry.Socket )
		
		If( A.ReadIncData() )
			Entry.WebServer.Func( A )
		EndIf

		A.Connection.Close()
	End Function
	
	'################################################
End Type

'##################################################################################################
Type TWebServerThreadEntry

	'################################################
	Field Socket:TSocket
	Field WebServer:TWebServer
	
	'################################################
	Function Create:TWebServerThreadEntry( Socket:TSocket, WebServer:TWebServer )
		Local A:TWebServerThreadEntry = New TWebServerThreadEntry
		A.Socket = Socket
		A.WebServer = WebServer
		Return( A )
	End Function
	
	'################################################
End Type

'##################################################################################################